#include <linux/version.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/ctype.h>

#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/smp_lock.h>

#include "ftpfs.h"
#include "ftpfs_proc.h"

static int ftp_lookup_validate(struct dentry*, int);
static int ftp_hash_dentry(struct dentry*, struct qstr*);
static int ftp_compare_dentry(struct dentry*, struct qstr*, struct qstr*);
static int ftp_delete_dentry(struct dentry*);

static struct dentry_operations ftpfs_dentry_operations = {
	d_revalidate:	ftp_lookup_validate,
	d_hash:		ftp_hash_dentry,
	d_compare:	ftp_compare_dentry,
	d_delete:	ftp_delete_dentry,
};

static int ftp_readdir(struct file*, void*, filldir_t);
static int ftp_dir_open(struct inode*, struct file*);

static struct dentry *ftp_lookup(struct inode*, struct dentry*);

struct file_operations ftp_dir_operations =  {
	read:		generic_read_dir,
	readdir:	ftp_readdir,
	open:		ftp_dir_open,
};

struct inode_operations ftp_dir_inode_operations = {
	lookup:		ftp_lookup,
};

static int
ftp_lookup_validate(struct dentry* dentry, int flags){
	struct inode *inode = dentry->d_inode;
	int valid = 1;

	DEBUG("\n");
	if(inode){
		lock_kernel();
		if(is_bad_inode(inode))
			valid = 0;
		unlock_kernel();
	}

	return valid;
}

static int
ftp_hash_dentry(struct dentry* dentry, struct qstr* this){
	unsigned long hash;
	int i;

	DEBUG("\n");

	hash = init_name_hash();
	for(i = 0; i < this->len; i++)
		hash = partial_name_hash(tolower(this->name[i]), hash);
	this->hash = end_name_hash(hash);
	return 0;
}

static int
ftp_compare_dentry(struct dentry* dentry, struct qstr* a, struct qstr* b){
	int i, res = 1;

	DEBUG("\n");
	if(a->len != b->len)
		goto out;
	for(i = 0; i < a->len; i++)
		if(tolower(a->name[i]) != tolower(b->name[i]))
			goto out;
	res = 0;
out:
	return res;
}

static int
ftp_delete_dentry(struct dentry* dentry){

	DEBUG("\n");
	if(dentry->d_inode){
		if(is_bad_inode(dentry->d_inode))
			return 1;
	}
	
	return 0;
}

static int
ftp_readdir(struct file* f , void* dirent, filldir_t filldir){
	struct dentry *dentry = f->f_dentry;
	struct inode *inode = dentry->d_inode;
	struct super_block *sb = inode->i_sb;
	struct ftp_sb_info *info = (struct ftp_sb_info*)sb->u.generic_sbp;
	char buf[FTP_MAXPATHLEN];
	int result, pos;

	struct ftp_directory *dir;
	struct ftp_dirlist_node *file;
	
	DEBUG(" reading %s, f_pos=%d\n", dentry->d_name.name, (int)f->f_pos);
	if(ftp_get_name(dentry, buf) < 0){
		VERBOSE(" ftp_get_name failed!\n");
		return -1;
	}
	
	result = 0;
	switch((unsigned int) f->f_pos){
		case 0:
			if(filldir(dirent, ".", 1, 0, inode->i_ino, DT_DIR) < 0)
				goto out;
			f->f_pos = 1;
		case 1:
			if(filldir(dirent, "..", 2, 1, dentry->d_parent->d_inode->i_ino, DT_DIR) < 0)
				goto out;
			f->f_pos = 2;
		default:
			ftp_lock(info);
			dir = ftp_cache_get(info, buf);
			ftp_unlock(info);
			if(!dir){
				VERBOSE(" ftp_cache_get failed!\n");
				return -1;
			}
//			for(pos = 0, file = dir->head; (pos < f->f_pos)&&(file != NULL); file = file->next); 
			pos = 2;
			for(file = dir->head; file != NULL; file = file->next){
//				DEBUG("pos=%u, f->f_pos=%u, name=%s\n", pos, f->f_pos, file->entry.name);
				if((pos == f->f_pos)){
					struct qstr qname;
					unsigned long ino;

					qname.name = file->entry.name;
					qname.len = strlen(qname.name);
//					DEBUG("pos=%u, f->f_pos=%u, name=%s\n", pos, f->f_pos, file->entry.name);

					ino = find_inode_number(dentry, &qname);
					if(!ino)
						ino = iunique(dentry->d_sb, 2);
					if(filldir(dirent, qname.name, qname.len, f->f_pos, ino, DT_UNKNOWN) >= 0)
						f->f_pos++;
					
				}
				pos++;
			}

	}

out:
	return result;
}

static int
ftp_dir_open(struct inode* inode, struct file* f){
	DEBUG("\n");
	return 0;
}


static struct dentry*
ftp_lookup(struct inode* dir, struct dentry* dentry){
	struct ftp_fattr fattr;
	struct inode *inode;
	
	DEBUG(" dname:%s\n", dentry->d_name.name);

	if(ftp_get_attr(dentry, &fattr, (struct ftp_sb_info*)dir->i_sb->u.generic_sbp) < 0){
		VERBOSE(" file not found!\n");
		return ERR_PTR(-ENOENT); 
	}

	fattr.f_ino = iunique(dentry->d_sb, 2);
	inode = ftp_iget(dir->i_sb, &fattr);
	if(inode){
		dentry->d_op = &ftpfs_dentry_operations;
		d_add(dentry, inode);
	}
	
	return NULL; 
}
